#include <stdint.h>
#include <stdio.h>

#define REF_CURRENT           50   // measured = Iin/100/2.5*1024 or target = REF_CURRENT/1024*2.5*100 :- 50 = 12.5A
#define REGULATOR_SCALE_SHIFT   6       /// Kp, and Ki are multiplied by 2^6

struct pwm_params {
	int16_t pwm_h;
	int16_t pwm_l;
	int16_t peak;
} pwm_out;

/**
* Function is called at A/D Sample Rate, and accumulates 64 samples, then
* executes simple PI regulator and outputs an 8-bit PWM.
*/
struct pwm_params buck_controller(uint16_t sample, int16_t Kp, int16_t Ki, uint8_t reset) {
    static uint8_t ad_sample_count = 0;
    static int32_t sample_acc = 0;
    static int16_t integ = 0;

    pwm_out.peak = 160; //50kHz @ 16MHz

    if (reset) {
        fprintf(stderr, "Firmware Reset\n"); //start kicad in a terminal to get this.
        ad_sample_count = 0;
        sample_acc = 0;
    }
    else {
        if (++ad_sample_count > 1) {
			sample_acc += sample;
            ad_sample_count = 0;
            sample_acc >>= 1;
            int32_t e = REF_CURRENT - sample_acc;
            integ += e * Ki;
            int32_t out = (e * Kp + integ * Ki) >> (REGULATOR_SCALE_SHIFT);
            pwm_out.pwm_h = (out > pwm_out.peak - 16) ? pwm_out.peak - 16 : (out < 16) ? 16 : out;
            pwm_out.pwm_l = pwm_out.pwm_h + 4; 
            
            fprintf(stderr, "I = %d, e = %d, integ = %d, PWMH = %d, PWML = %d\n", sample, e, integ, pwm_out.pwm_h, pwm_out.pwm_l); //start kicad in a terminal to get this.
            sample_acc = 0;
            //pwm_out.pwm_h = 70;
            //pwm_out.pwm_l = 80;
        }
        else {
            sample_acc += sample;
        }
    }
    return pwm_out;
}
